/*
 * This file is part of openSDK.
 *
 * Copyright (C) 2006-2007 openSDK team
 *
 * openSDK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * openSDK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with openSDK; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *     $Id: OCalendarTime.cc,v 1.8 2007/02/14 17:34:38 nuno-lopes Exp $
 */

#include <OPENR/OCalendarTime.h>
#include <cstdio>
#include <cstring>
#include <time.h>
#include "Platform.h"

// we cannot use strftime() because of non-C locals
static const char monthNames[][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
static const char DOWNames[][4]   = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"}; // sunday starts at 0


static OStatus convert_to_calendar(OCalendarTime& calendar, const OTime& time)
{
	struct tm timeStruct;
	time_t clock = time.GetClock() + TIME_DIFF_SYSTEM;

	if (gmtime_r(&clock, &timeStruct) == NULL){
		return oFAIL;
	}

	calendar.year_    = timeStruct.tm_year + 1900 - otimeSTART_YEAR;
	calendar.month_   = timeStruct.tm_mon + 1;
	calendar.day_     = timeStruct.tm_mday;
	calendar.hour_    = timeStruct.tm_hour;
	calendar.min_     = timeStruct.tm_min;
	calendar.sec_     = timeStruct.tm_sec;
	calendar.dow_     = timeStruct.tm_wday;
	calendar.timeDif_ = time.GetTimeDif();

	return oSUCCESS;
}

static const time_t convert_to_timestamp(const OCalendarTime& calendar)
{
	time_t clock;
	sbyte timeDif;
	calendar.GetTime(&clock, &timeDif);
	return clock + (timeDif * 3600);
}


OCalendarTime::OCalendarTime()
{
	OCalendarTime(0, 0);
}

OCalendarTime::OCalendarTime(time_t clock, sbyte timeDif)
{
	OTime time(clock, timeDif);
	convert_to_calendar(*this, time);
}

OCalendarTime::OCalendarTime(const OTime& time)
{
	convert_to_calendar(*this, time);
}

OCalendarTime::OCalendarTime(const OCalendarTime& calendar)
{
	*this = calendar;
}

OCalendarTime& OCalendarTime::operator=(const OCalendarTime& calendar)
{
	year_    = calendar.year_;
	month_   = calendar.month_;
	day_     = calendar.day_;
	hour_    = calendar.hour_;
	min_     = calendar.min_;
	sec_     = calendar.sec_;
	dow_     = calendar.dow_;
	timeDif_ = calendar.timeDif_;

	return *this;
}

bool OCalendarTime::operator<(const OCalendarTime& calendar) const
{
	return (convert_to_timestamp(*this) < convert_to_timestamp(calendar));
}

bool OCalendarTime::operator<=(const OCalendarTime& calendar) const
{
	return (convert_to_timestamp(*this) <= convert_to_timestamp(calendar));
}

bool OCalendarTime::operator>(const OCalendarTime& calendar) const
{
	return (convert_to_timestamp(*this) > convert_to_timestamp(calendar));
}

bool OCalendarTime::operator>=(const OCalendarTime& calendar) const
{
	return (convert_to_timestamp(*this) >= convert_to_timestamp(calendar));
}

bool OCalendarTime::operator==(const OCalendarTime& calendar) const
{
	return (convert_to_timestamp(*this) == convert_to_timestamp(calendar));
}

OStatus OCalendarTime::SetTime(time_t clock, sbyte timeDif)
{
	OTime time(clock, timeDif);
	return convert_to_calendar(*this, time);
}

OStatus OCalendarTime::SetTime(const OTime& time)
{
	return convert_to_calendar(*this, time);
}

OStatus OCalendarTime::GetTime(time_t* clock, sbyte* timeDif) const
{
	struct tm timeStruct;
	timeStruct.tm_year   = year_ + otimeSTART_YEAR - 1900;
	timeStruct.tm_mon    = month_ - 1;
	timeStruct.tm_mday   = day_;
	timeStruct.tm_hour   = hour_;
	timeStruct.tm_min    = min_;
	timeStruct.tm_sec    = sec_;
	timeStruct.tm_isdst  = 0;

	if ((*clock = mktime(&timeStruct)) == -1) {
		return oFAIL;
	}

	*timeDif = timeDif_;
	*clock -= TIME_DIFF_SYSTEM;
	return oSUCCESS;
}

OStatus OCalendarTime::GetTime(OTime* time) const
{
	return GetTime(&(time->clock_), &(time->timeDif_));
}

bool OCalendarTime::ProperDate() const
{
	OTime dummy;
	return (GetTime(&dummy) == oSUCCESS ? true : false);
}

OStatus OCalendarTime::CorrectDayOfTheWeek()
{
	OTime dummy;
	return ((GetTime(&dummy) == oSUCCESS && convert_to_calendar(*this, dummy) == oSUCCESS) ? oSUCCESS : oFAIL);
}

void OCalendarTime::Print() const
{
	// TODO
}

void OCalendarTime::Sprint(char* str) const
{
	RFC2822(str);
}

void OCalendarTime::RFC2822(char* str) const
{
	if (dow_ < 0 || dow_ > 6) {
		strcpy(str, "OCalendarTime::RFC2822() : Invalid day of week");
	} else if (month_ < 1 || month_ > 12) {
		strcpy(str, "OCalendarTime::RFC2822() : Invalid month");
	} else {
		sprintf(str, "%s, %02d %s %d %02d:%02d:%02d %+02d00", DOWNames[dow_], day_, monthNames[month_-1], year_+otimeSTART_YEAR, hour_, min_, sec_, timeDif_);
	}
}
